import dataclasses
import os
import signal

import psutil

from aica_challenge_1 import Challenge, RunSpecification
from pathlib import Path
from pythonclimenu import menu
from typing import Dict, List, Callable, Any, Optional

from aica_challenge_1.execution_manager import RunStatus, Run

background_color = (100, 100, 100)

def wait_for_key():
    print("Press any key to continue...")
    input()

# Forward references, yay!
menus: Dict[str, Any] = {}

menu_top_level = {
    "back": None,
    "menu": [
        f"AICA Challenge 1",
        {
            "Packages/Agents": None,
            "Scenarios": None,
            "Execution": None,
            "Exit": None
        },
        background_color
    ]}

def list_installed_agents(challenge: Challenge):
    print("Installed agents:")
    for agent in challenge.packages.list_installed_agents().values():
        print(f"Agent: {agent.name}, Package: {agent.package}, Description: {agent.description}")
    print("")
    wait_for_key()

def list_available_packages(challenge: Challenge):
    print("Available packages:")
    for package in challenge.packages.list_available_packages():
        print(package)
    print("")
    wait_for_key()

def install_package(challenge: Challenge, package_name):
    print(f"Installing a package from the directory '{package_name}'...", end='', flush=True)
    success, out, err = challenge.packages.install_package(package_name)
    if success:
        print("[OK]")
    else:
        print("[ERROR]")
        print(out.replace("\\n", '\n').replace('\\t', '\t').replace('\\r', '\r'))
        print(err.replace("\\n", '\n').replace('\\t', '\t').replace('\\r', '\r'))

    print("")

def install_package_menu(challenge: Challenge):
    available_packages = challenge.packages.list_available_packages()
    if not available_packages:
        print("No packages available to install")
        wait_for_key()
        return

    available_packages.append("Cancel")
    choice = menu("Install package: ", available_packages, background_color)
    if choice != "Cancel":
        install_package(challenge, choice)
        list_installed_agents(challenge)

def install_all(challenge: Challenge):
    available_packages = challenge.packages.list_available_packages()
    if not available_packages:
        print("No packages available to install")
        wait_for_key()
    else:
        for package in available_packages:
            install_package(challenge, package)
        list_installed_agents(challenge)

def remove_package(challenge: Challenge, package_name: str):
    print(f"Removing the package '{package_name}'...", end='', flush=True)
    success, out, err = challenge.packages.remove_package(package_name)
    if success:
        print("[OK]")
    else:
        print("[ERROR]")
        print(out)
        print(err)

    print("")

def remove_package_menu(challenge: Challenge):
    installed_packages = [x.name for x in challenge.packages.list_installed_packages()]
    if not installed_packages:
        print("No packages available to remove")
        wait_for_key()
        return

    installed_packages.append("Cancel")
    choice = menu("Remove package: ", installed_packages, background_color)
    if choice != "Cancel":
        remove_package(challenge, choice)
        list_installed_agents(challenge)

def remove_all(challenge: Challenge):
    installed_packages = [x.name for x in challenge.packages.list_installed_packages()]
    if not installed_packages:
        print("No packages available to remove")
        wait_for_key()
    else:
        for package in installed_packages:
            remove_package(challenge, package)
        list_installed_agents(challenge)

def init_agent(challenge: Challenge):
    agent_name = input("Please provide a name for the new agent: ")
    result, normalized_name = challenge.packages.init_agent(agent_name)

    if not result:
        wait_for_key()
        return

    choice = menu((f"Agent '{normalized_name}' was successfully initialized. Would you like to install it right away?", "(You can do it later without any problem)"),
                  ["Yes", "No"],
                  background_color)

    if choice == "Yes":
        install_package(challenge, normalized_name)
        list_installed_agents(challenge)

menu_packages = {
    "back": menu_top_level,
    "menu": [
        "Package/Agent management",
        {
            "List installed agents": list_installed_agents,
            "List available packages": list_available_packages,
            "Install package": install_package_menu,
            "Install all": install_all,
            "Remove package": remove_package_menu,
            "Remove all": remove_all,
            "Init agent": init_agent,
            "Back": None
        },
        background_color
    ]}

def list_remote_scenarios(challenge: Challenge):
    print("Remotely available scenarios:")
    for scenario in challenge.scenarios.get_scenarios(remote=True):
        print(f"\nName: {scenario.name}\nDescription: {scenario.description}\nGoal: {scenario.goal}\nVariants: {[x.id for x in scenario.variants.values()]}")
    print("")
    wait_for_key()

def list_local_scenarios(challenge: Challenge):
    print("Locally available scenarios:")
    for scenario in challenge.scenarios.get_scenarios():
        print(f"\nName: {scenario.name}\nDescription: {scenario.description}\nGoal: {scenario.goal}\nVariants: {sorted([x.id for x in scenario.variants.values()])}")
    print("")
    wait_for_key()

def download_scenarios(challenge: Challenge) -> None:
    choice = menu("Do you want to overwrite local files with remote ones?",
                  ["Yes", "No", "Back"],
                  background_color)

    if choice == "Back":
        return

    challenge.scenarios.download_remote_scenarios(overwrite=(choice == "Yes"))
    print("Scenarios successfully downloaded")
    wait_for_key()

menu_scenarios = {
    "back": menu_top_level,
    "menu": [
        "Scenario management",
        {
            "List remote scenarios": list_remote_scenarios,
            "List local scenarios": list_local_scenarios,
            "Download scenarios": download_scenarios,
            "Back": None
        },
        background_color
]}

def choose_agent(challenge: Challenge) -> Optional[str]:
    options = list(challenge.packages.list_installed_agents().keys())
    options.append("Back")

    choice = menu("Select agent:",
                  options,
                  background_color)

    if choice == "Back":
        return None
    else:
        return choice

def choose_scenario(challenge: Challenge) -> Optional[str]:
    available_scenarios = {x.name: x for x in challenge.scenarios.get_scenarios()}
    options = [x.name for x in available_scenarios.values()]
    options.append("Random")
    options.append("Back")

    choice = menu("Choose, which scenario to use in runs:",
                  options,
                  background_color)

    if choice == "Back":
        return None
    elif choice == "Random":
        return "Random"
    else:
        return available_scenarios[choice].short_path

def choose_variant(challenge: Challenge, scenario: str) -> Optional[int]:
    if scenario == "Random":
        print("Random scenario supports only random variants.")
        wait_for_key()
        return -1

    available_variants = challenge.scenarios.get_scenario(scenario).variants

    error = ""
    while True:
        print(f"Variants available for scenario '{scenario}': {sorted(available_variants.keys())}")
        if error:
            print(error)

        v = input("Choose, which variants to use or type '-1' for random variant: ")
        try:
            v_int = int(v)
        except:
            error = "Please specify a number"
            continue

        if v_int != -1 and v_int not in available_variants:
            error = "Please select one of the variants"
        else:
            return v_int

def set_run_parameters(challenge: Challenge, template: RunSpecification) -> None:
    parameters = iter(["max_time", "max_actions", "max_episodes", "max_parallel"])

    next_parameter = next(parameters)
    while True:
        user_input = input(f"Enter new value for '{next_parameter}'. (Current value - {getattr(template, next_parameter)}): ")

        try:
            value = int(user_input)
        except ValueError:
            print("Please, enter a number")
            continue

        setattr(template, next_parameter, value)
        next_parameter = next(parameters, None)
        if not next_parameter:
            break

def set_custom_parameters(challenge: Challenge, template: RunSpecification) -> None:
    if template.parameters:
        print("Modifying already set parameters")
        for name, value in template.parameters.items():
            user_input = input(f"Enter new value for '{name}'. (Current value - '{value}'): ")
            template.parameters[name] = user_input

    print("Adding new parameters (empty parameter name to end)")
    while True:
        name = input("Enter parameter name: ")
        if not name:
            break
        elif name in template.parameters:
            print(f"Parameter with the name '{name}' already exists. Ignoring...")
            continue
        else:
            value = input("Enter parameter value: ")
        template.parameters[name] = value

def view_template(challenge: Challenge, template: RunSpecification) -> None:
    print(f"{'Name: ':<37}{template.name}")
    print(f"{'Description: ':<37}{template.description}")
    print(f"{'Agent: ':<37}{template.agent_name}")
    scenario_str = "Random" if template.scenario == "Random" else challenge.scenarios.get_scenario(template.scenario).name
    print(f"{'Scenario: ':<37}{scenario_str}")
    variant_str = "Random" if template.variant == -1 else str(template.variant)
    print(f"{'Variant: ':<37}{variant_str}")
    print(f"{'Maximum number of episodes: ':<37}{template.max_episodes}")
    print(f"{'Maximum time per episode: ':<37}{template.max_time}")
    print(f"{'Maximum actions per episode: ':<37}{template.max_actions}")
    print(f"{'Maximum number of parallel episodes: ':<37}{template.max_parallel}")
    print(f"Custom parameters: ", end="")
    if not template.parameters:
        print("                  None")
    else:
        print("")
        for name, value in template.parameters.items():
            print(f"    {name}:{value}")

def modify_template(challenge: Challenge, template: RunSpecification) -> RunSpecification:
    old_template = RunSpecification.copy(template)
    old_template.name = template.name

    if template.name:
        menu_title = f"Modifying a run template '{template.name}'"
    else:
        menu_title = "Creating a new run template"

    while True:
        choice = menu(menu_title,
                      ["Set name", "Set description", "Set agent", "Set scenario(s)", "Set variant(s)",
                              "Set run parameters", "Set custom parameters", "View template", "Save", "Back"],
                      background_color)

        if choice == "Back":
            break

        elif choice == "Set name":
            if template.name:
                print(f"Current template name: {template.name}")
            name = input("Please enter new name (or just hit enter to keep it): ")
            if name:
                template.name = name
                menu_title = f"Modifying a run template '{template.name}'"

        elif choice == "Set description":
            if template.description:
                print(f"Current template description: {template.description}")
            description = input("Please enter new description (or just hit enter to keep it): ")
            if description:
                template.description = description

        elif choice == "Set agent":
            agent_name = choose_agent(challenge)
            if agent_name:
                template.agent_name = agent_name

        elif choice == "Set scenario(s)":
            scenario = choose_scenario(challenge)
            if scenario:
                template.scenario = scenario

        elif choice == "Set variant(s)":
            variant = choose_variant(challenge, str(template.scenario))
            if variant:
                template.variant = variant

        elif choice == "Set run parameters":
            set_run_parameters(challenge, template)
            view_template(challenge, template)
            wait_for_key()

        elif choice == "Set custom parameters":
            set_custom_parameters(challenge, template)
            view_template(challenge, template)
            wait_for_key()

        elif choice == "View template":
            view_template(challenge, template)
            wait_for_key()

        elif choice == "Save":
            if not template.name:
                print("Cannot save a template without a name")
                wait_for_key()
                continue

            tmp = challenge.execution.get_run_specification(template.name)
            save = True
            if tmp:
                yes_no = menu(f"Do you want to overwrite the template '{template.name}",
                              ["Yes", "No"],
                              background_color)
                if yes_no == "No":
                    save = False

            if save:
                challenge.execution.set_run_specification(template, old_template)
                print(f"Run specification '{template.name}' saved.")
                wait_for_key()

    return template

def get_run_template(challenge: Challenge) -> RunSpecification:
    available_templates = challenge.execution.list_run_specifications()

    if not available_templates:
        print("No templates are available.")
        wait_for_key()
        return RunSpecification()

    available_templates.append("None")

    choice = menu("Select a run specification to use as a template:",
                  available_templates,
                  background_color)

    if choice == "None":
        return RunSpecification()
    else:
        return challenge.execution.get_run_specification(choice)

def new_run_specification(challenge: Challenge) -> None:
    choice = menu("Do you want to modify a saved run template?",
                  ["Yes", "No", "Back"],
                  background_color)

    if choice == "Back":
        return
    elif choice == "Yes":
        run_template = get_run_template(challenge)
        if not run_template.name:
            template = RunSpecification()
        else:
            template = RunSpecification.copy(run_template)
    else:
        template = RunSpecification()

    modify_template(challenge, template)

def edit_run_specification(challenge: Challenge) -> None:
    spec = get_run_template(challenge)
    if not spec.name:
        return
    else:
        modify_template(challenge, spec)

def new_run(challenge: Challenge) -> None:
    available_runs = challenge.execution.list_run_specifications()
    available_runs.append("Back")
    while True:
        choice = menu("Which run would you like to start?",
                      available_runs,
                      background_color)

        if choice == "Back":
            return

        run_spec = challenge.execution.get_run_specification(choice)
        choice_2 = menu(f"You have selected run '{choice}' on scenario '{run_spec.scenario}' with agent '{run_spec.agent_name}. Is this correct?",
                        ["Yes", "No", "Back"],
                        background_color)

        if choice_2 == "Back":
            continue
        elif choice_2 == "No":
            break
        else:
            challenge.execution.execute(run_spec)
            break


def show_run_status(run: Run):
    print(f"Run id: {run.id}")
    print(f"Status: {run.status}")
    print(f"Episodes:")
    print(f"  - All     : {sorted(run.episodes.keys())}")
    print(f"  - Running : {sorted(run.running)}")
    print(f"  - Finished: {sorted(run.successful)}")
    print(f"  - Error   : {sorted(run.error)}")
    wait_for_key()


def status(challenge: Challenge):
    categorized_runs = {
        RunStatus.INIT: [],
        RunStatus.RUNNING: [],
        RunStatus.FINISHED: [],
        RunStatus.ERROR: []
    }

    all_runs = challenge.execution.get_runs()
    for run in all_runs:
        categorized_runs[run[1]].append(run[0])

    print("Status of runs (status: IDs)\n")
    print(f"{'Initialized: ':<14}{"None" if not categorized_runs[RunStatus.INIT] else categorized_runs[RunStatus.INIT]}")
    print(f"{'Running: ':<14}{"None" if not categorized_runs[RunStatus.RUNNING] else categorized_runs[RunStatus.RUNNING]}")
    print(f"{'Finished: ':<14}{"None" if not categorized_runs[RunStatus.FINISHED] else categorized_runs[RunStatus.FINISHED]}")
    print(f"{'Error: ':<14}{"None" if not categorized_runs[RunStatus.ERROR] else categorized_runs[RunStatus.ERROR]}")
    print("")

    while True:
        choice = input("Do you want to view a detailed status of any run (just press enter to return back)? ")
        if not choice:
            break

        try:
            run_id = int(choice)
        except:
            print(f"Selected value '{choice}' does not convert to integral run id. Please, try again.")
            continue

        if run_id not in [x[0] for x in all_runs]:
            print(f"Selected run id '{run_id}' not among the runs. Please, try again.")
            continue

        show_run_status(challenge.execution.get_run(run_id))
        break


def evaluation(challenge: Challenge):
    print("Evaluation is currently disabled")
    wait_for_key()

menu_execution = {
    "back": menu_top_level,
    "menu": [
        "Execution of training and evaluation runs",
        {
            "New run specification": new_run_specification,
            "Edit run specification": edit_run_specification,
            "New run": new_run,
            "Run status": status,
            "New evaluation": evaluation,
            "Back": None
        },
        background_color
]}

menus["Packages/Agents"] = menu_packages
menus["Scenarios"] = menu_scenarios
menus["Execution"] = menu_execution

def run():
    try:
        challenge = Challenge()
    except ValueError as e:
        path = e.args[1]
        choice = menu(("The challenge environment does not appear to be initiated in the current running directory.", str(path), "Would you like to..."),
                      ["Init the challenge environment here", "Set the path to the challenge environment", "Exit and run this from elsewhere"],
                      background_color)

        if choice.startswith("Init"):
            Challenge.init_environment()
            challenge = Challenge()
        elif choice.startswith("Exit"):
            return
        else:
            path_set_correctly = False
            while not path_set_correctly:
                path_str = input("Please, provide the path to the environment: ")
                path = Path(path_str)
                if not Challenge.check_init_state(path):
                    print("Provided directory is not a correct challenge environment.")
                else:
                    path_set_correctly = True
                    challenge = Challenge(str(path))

    next_menu = menu_top_level

    while True:
        menu_spec = next_menu["menu"]
        choice = menu(menu_spec[0], list(menu_spec[1].keys()), menu_spec[2])
        if choice == "Back":
            next_menu = next_menu["back"]
            continue
        elif choice == "Exit":
            all_runs = challenge.execution.get_runs()
            if RunStatus.RUNNING in [x[1] for x in all_runs]:
                choice = menu("Some runs have not finished. If you exit, it will terminate them. Do you really want to exit?",
                              ["Yes", "No"], background_color)
                if choice == "No":
                    continue
                else:
                    current_process = psutil.Process()
                    children_process = current_process.children(recursive=True)
                    for p in children_process:
                        try:
                            p.send_signal(signal.SIGTERM)
                        except psutil.NoSuchProcess:
                            pass
                    break
            break
        else:
            selected = menu_spec[1][choice]
            if not selected:
                next_menu = menus[choice]
            elif isinstance(selected, Callable):
                selected(challenge)


if __name__ == "__main__":
    run()



import importlib
import importlib.metadata
import pkgutil
import re
import subprocess
import sys
import zipfile

from dataclasses import dataclass, field
from pathlib import Path
from string import Template
from typing import Optional, List, Tuple, Set, Dict

from cyst.api.host.service import ActiveServiceDescription


@dataclass
class InstalledAgent:
    """
    A dataclass tracking information about an installed challenge-compatible agent.

    :param name: A unique name of the agent.
    :param package: A name of the package which the agent is a part of. Multiple agents can be in one package.
    :param description: A description of an agent.
    """
    name: str
    package: str
    description: str


@dataclass
class InstalledPackage:
    """
    A dataclass tracking information about installed packages that contain challenge-compatible agents.

    :param name: A name of the package.
    :param path: A full path to the package as returned by the module_finder.
    :param dir: A directory parth of the path.
    :param version: A version of the package.
    :param description: A description of the package.
    """
    name: str
    path: str
    dir: str
    version: str
    description: str

    def __lt__(self, other):
        return self.dir < other.dir


@dataclass
class PackageList:
    installed: List[InstalledPackage] = field(default_factory=lambda: [])
    available: List[str] = field(default_factory=lambda: [])


class PackageManager:
    """
    Package manager provides convenience functions for initializing, installing and removing agents. Be aware that these
    are really just convenience functions and rely on the underlying package management system (i.e., Poetry). Therefore
    if you have an agent that exists as a separate package somewhere
    (e.g., <https://pypi.org/project/aica-challenge-1-heuristic-agent/>),
    you can simply install it via poetry and it
    will work as if you used this package manager.
    """
    def __init__(self, agents_path: str | Path = ""):
        if not agents_path:
            self._package_path: Path = Path(__file__).parent / "agents"
        elif isinstance(agents_path, str):
            self._package_path: Path = Path(agents_path)
        else:
            self._package_path = agents_path

    @staticmethod
    def list_installed_agents() -> Dict[str, InstalledAgent]:
        """
        Provides a list of challenge-compatible agents that are installed in the current virtual environment.

        :return: A dictionary, where the key is an agent name and the value is the agent description structure.
        """
        ignored_services = ["firewall", "forward_shell", "reverse_shell", "scripted_actor"]
        result = {}
        for service in importlib.metadata.entry_points(group="cyst.services"):
            if service.name in ignored_services:
                continue
            service_description: ActiveServiceDescription = service.load()
            result[service.name] = InstalledAgent(
                service.name,
                service.dist.name,
                service_description.description
            )
        return result

    def check_packages(self) -> PackageList:
        result = PackageList()

        packages_on_path = set([x.parts[-1] for x in self._package_path.iterdir() if x.is_dir()])
        installed_packages = set()

        for module in pkgutil.iter_modules():
            if hasattr(module.module_finder, "path"):
                module_path = Path(module.module_finder.path)
                module_dir = module_path.parts[-2]
                if module_path.is_relative_to(self._package_path):
                    installed_packages.add(module_dir)
                    metadata = importlib.metadata.metadata(module.name)
                    result.installed.append(InstalledPackage(
                        metadata.get("Name"),
                        str(module_path),
                        module_dir,
                        metadata.get("Version"),
                        metadata.get("Summary")
                    ))

        result.available = list(packages_on_path - installed_packages)

        return result

    def list_installed_packages(self) -> List[InstalledPackage]:
        """
        Provides a list of packages installed in the current virtual environment from the `agents` directory of the
        challenge.

        :return: A list of installed packages.
        """
        return self.check_packages().installed

    def list_available_packages(self) -> List[str]:
        """
        Provides a list of packages that are available for installation in the `agents` directory of the challenge.

        :return: A list of available packages.
        """
        return self.check_packages().available

    def install_package(self, name: str) -> Tuple[bool, str, str]:
        """
        Attempts to install a package from the given path.

        :param name: A name of the package that should be the same as a name of a directory in the `agents` folder
            of the challenge. While you can escape the directory with '..' I suggest not to do it.

        :return: A tuple indicating whether the installation was successful [0] and if not, also providing the
            stdout [1] and stderr [2].
        """
        for installed_package in self.list_installed_packages():
            if name == installed_package.dir:
                return True, "", ""

        result = subprocess.run(["poetry", "add", "-e", str(self._package_path / name)], capture_output=True)
        if result.returncode == 0:
            sys.path.append(str(self._package_path / name / "src"))
            return True, "", ""
        else:
            return False, str(result.stdout), str(result.stderr)

    def install_all(self) -> bool:
        """
        Attempts to install all available packages in the `agents` directory.

        :return: True on success, False otherwise.
        """
        failure = False
        for available_package in sorted(self.list_available_packages()):
            failure |= self.install_package(available_package)
        return not failure

    def remove_package(self, name: str) -> Tuple[bool, str, str]:
        """
        Attempts to remove a package by name. Be aware that this has to be a package name and not a path to the package.
        As this is just a thin wrapper over poetry, it will happily remove any package in the current virtual
        environment.

        :param name: A name of the package to remove.

        :return: A tuple indicating whether the removal was successful [0] and if not, also providing the
            stdout [1] and stderr [2].
        """
        for installed_package in self.list_installed_packages():
            if name == installed_package.dir:
                result = subprocess.run(["poetry", "remove", installed_package.name], capture_output=True)
                if result.returncode == 0:
                    sys.path.remove(installed_package.path)
                    return True, "", ""
                else:
                    return False, str(result.stdout), str(result.stderr)
        return True, "", ""

    def remove_all(self) -> bool:
        """
        Attempts to remove all packages in the `agents` directory from the current virtual environment.

        :return: True on success, False otherwise.
        """
        failure = False
        for installed_agent in sorted(self.list_installed_packages()):
            failure |= self.remove_package(installed_agent.dir)
        return not failure

    def init_agent(self, agent_name: str) -> Tuple[bool, str]:
        """
        Initializes an agent's skeleton in the `agents` directory of the current challenge.

        :param agent_name: A desired name of the agent.

        :return: Tuple indicating whether the initialization was successful [0] and its normalized PEP-compliant name
            [1].
        """
        if not re.match("^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$", agent_name, re.IGNORECASE):
            print(f"Agent name '{agent_name}' does not conform to package name convention. See PEP 503 and 508")
            return False, agent_name

        normalized_name = re.sub(r"[-_.]+", "-", agent_name).lower()
        normalized_dir_name = re.sub(r"-+", "_", normalized_name)

        agent_path = Path(self._package_path / normalized_name)
        if agent_path.exists():
            print(f"Agent with name '{agent_name}' already present at '{agent_path}'. Init aborted.")
            return False, normalized_name

        print(f"Initiating an agent with normalized name: {normalized_name}")

        agent_path.mkdir()

        agent_zip = zipfile.ZipFile(Path(__file__).parent / "agent_template._zip")
        agent_zip.extractall(agent_path)

        # The template need a couple of renamings. We provide the template, so only the happy path is considered.
        # - directory in src
        # - name in pyproject.toml
        # - name in ActiveServiceDescription
        agent_src_dir = agent_path / "src" / "agent_name"
        agent_src_dir.rename(agent_path / "src" / normalized_dir_name)

        # The desire to one-line intensifies
        agent_pyproject = Template((agent_path / "pyproject.toml").read_text())
        (agent_path / "pyproject.toml").write_text(agent_pyproject.substitute(agent_name=normalized_name, agent_path=normalized_dir_name))

        agent_source = Template((agent_path / "src" / normalized_dir_name / "main.py").read_text())
        (agent_path / "src" / normalized_dir_name / "main.py").write_text(agent_source.substitute(agent_name=normalized_name))

        print(f"Agent '{normalized_name}' successfully initiated")

        return True, normalized_name


import io
import os

from contextlib import redirect_stderr, redirect_stdout, nullcontext
from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session
from typing import Dict, Any

from aica_challenge_1.scenario_manager import ScenarioVariant
from cyst.api.environment.environment import Environment

from aica_challenge_1.execution_structures import Episode, DBEpisode, DBRun, RunStatus

def launch_simulation(run_id: int, episode_number: int, scenario: ScenarioVariant, agent: str, max_time: float,
                      max_actions: int, agent_configuration: Dict[str, str] | None = None, supress_output=True) -> Episode:
    parameters: Dict[str, Any] = { "agent-name": agent }
    if agent_configuration:
        parameters["agent-configuration"] = agent_configuration

    os.environ["CYST_MAX_RUNNING_TIME"] = str(max_time)
    os.environ["CYST_MAX_ACTION_COUNT"] = str(max_actions)
    os.environ["CYST_DATA_BACKEND"] = "sqlite"
    os.environ["CYST_DATA_BACKEND_PARAMS"] = "path,aica_challenge.db"
    os.environ["CYST_RUN_ID_LOG_SUFFIX"] = "1"

    db = create_engine("sqlite+pysqlite:///aica_challenge.db")

    with redirect_stdout(io.StringIO()) if supress_output else nullcontext() as stdout:
        with redirect_stderr(io.StringIO()) if supress_output else nullcontext() as stderr:

            env = Environment.create()
            try:
                env.configure(*scenario.config, parameters=parameters)
            except RuntimeError as e:
                print(f"Could not configure a simulation. Reason: {str(e)}")

            env.control.init()

            episode_id = -1

            with Session(db) as session:
                stmt = select(DBRun).where(DBRun.id == run_id)
                run: DBRun = session.scalars(stmt).one()

                db_episode = DBEpisode(
                        stdout="",
                        stderr="",
                        cyst_run_id=env.infrastructure.statistics.run_id,
                        status=str(RunStatus.RUNNING),
                        run_id=run_id,
                        run=run,
                        episode_number=episode_number
                    )

                session.add(db_episode)
                run.episodes.append(db_episode)

                session.add(run)
                session.flush()

                episode_id = db_episode.id

                session.commit()

            success, state = env.control.run()
            env.control.commit()

            episode = Episode(
                cyst_run_id=env.infrastructure.statistics.run_id,
                stdout=stdout.getvalue() if supress_output else "",
                stderr=stderr.getvalue() if supress_output else "",
                run=run_id,
                status=RunStatus.FINISHED if success else RunStatus.ERROR
            )

            with Session(db) as session:
                stmt = select(DBEpisode).where(DBEpisode.id == episode_id)
                db_episode: DBEpisode = session.scalars(stmt).one()

                db_episode.stdout = episode.stdout
                db_episode.stderr = episode.stderr
                db_episode.status = str(episode.status)

                session.commit()

    return episode


from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass, field
from enum import Enum, auto
from sqlalchemy import ForeignKey
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, MappedAsDataclass, relationship
from typing import List, Optional, Dict, Set, Self


class Base(MappedAsDataclass, DeclarativeBase):
    pass


class DBRunSpecification(Base):
    __tablename__ = "challenge_run_specification"

    id: Mapped[int] = mapped_column(init=False, primary_key=True)
    name: Mapped[str] = mapped_column(init=True, default="")
    description: Mapped[str] = mapped_column(init=True, default="")
    agent_name: Mapped[str] = mapped_column(init=True, default="")
    scenario: Mapped[str] = mapped_column(init=True, default="Random")
    variant: Mapped[int] = mapped_column(init=True, default=-1)
    max_time: Mapped[int] = mapped_column(init=True, default=100)
    max_actions: Mapped[int] = mapped_column(init=True, default=100)
    max_episodes: Mapped[int] = mapped_column(init=True, default=100)
    max_parallel: Mapped[int] = mapped_column(init=True, default=1)

    @staticmethod
    def copy(other: 'RunSpecification'):
        return DBRunSpecification(other.name, other.description, other.agent_name, other.scenario, other.variant,
                                other.max_time, other.max_episodes, other.max_parallel)

    def replace(self, other: 'RunSpecification'):
        self.name = other.name
        self.description = other.description
        self.agent_name = other.agent_name
        self.scenario = other.scenario
        self.variant = other.variant
        self.max_time = other.max_time
        self.max_actions = other.max_actions
        self.max_episodes = other.max_episodes
        self.max_parallel = other.max_parallel

@dataclass
class RunSpecification:
    """
    A specification of a run to be executed.

    :param id: A unique id of the run. It is autogenerated.
    :type id: int

    :param name: A name of the run. Must be unique, or the original gets overwritten.
    :type name: str

    :param description: A description of the run.
    :type description: str

    :param agent_name: A name of the agent that should participate in the run. It must be the agent's name, not a
        package name.
    :type agent_name: str

    :param scenario: A name of the scenario for the run. By default, it is 'Random'
    :type scenario: str

    :param variant: An identifier of the scenario variant. By default, it is '-1', which means random.
    :type variant: int

    :param max_time: A maximum number of virtual seconds that can elapse before the run is terminated.
    :type max_time: int

    :param max_actions: A maximum number of actions from one agent before the run is terminated.
    :type max_actions: int

    :param max_episodes: A number of episodes that should be executed. If the scenario or variant are random, they will
        be randomly selected for each episode.
    :type max_episodes: int

    :param max_parallel: A maximum number of episodes running in parallel. Ignored if executed with the single_process
        flag.
    :type max_parallel: int

    :param parameters: Other, user or system-specified parameters.
    :type parameters: Dict[str, str]
    """
    name: str = ""
    description: str = ""
    agent_name: str = ""
    scenario: str = "Random"
    variant: int = -1
    max_time: int = 100
    max_actions: int = 100
    max_episodes: int = 100
    max_parallel: int = 1
    parameters: Dict[str, str] = field(default_factory=dict)
    db_id: int = -1

    @staticmethod
    def from_db_spec(db_spec: DBRunSpecification):
        return RunSpecification(
            db_spec.name,
            db_spec.description,
            db_spec.agent_name,
            db_spec.scenario,
            db_spec.variant,
            db_spec.max_time,
            db_spec.max_actions,
            db_spec.max_episodes,
            db_spec.max_parallel,
            {},
            db_spec.id
        )

    @staticmethod
    def copy(other: 'RunSpecification'):
        parameters = {}
        for k, v in other.parameters.items():
            parameters[k] = v

        return RunSpecification(
            other.name + '_copy',
            other.description,
            other.agent_name,
            other.scenario,
            other.variant,
            other.max_time,
            other.max_actions,
            other.max_episodes,
            other.max_parallel,
            parameters,
            other.db_id
        )

class DBRunParameter(Base):
    __tablename__ = "challenge_run_parameters"

    id: Mapped[int] = mapped_column(init=False, primary_key=True)
    run_id: Mapped[int] = mapped_column(init=True)
    key: Mapped[str] = mapped_column(init=True)
    value: Mapped[str] = mapped_column(init=True)


class DBRun(Base):
    __tablename__ = "challenge_run_statistics"

    id: Mapped[int] = mapped_column(init=False, primary_key=True)
    status: Mapped[str] = mapped_column(init=True)
    details: Mapped[str] = mapped_column(init=True)
    episodes: Mapped[List['DBEpisode']] = relationship(argument="DBEpisode",
                                                       back_populates="run",
                                                       cascade="all, delete")

    specification_id: Mapped[int] = mapped_column(ForeignKey("challenge_run_specification.id"), init=True)
    specification: Mapped[DBRunSpecification] = relationship()


class DBEpisode(Base):
    __tablename__ = "challenge_episode_statistics"

    id: Mapped[int] = mapped_column(init=False, primary_key=True)
    episode_number: Mapped[int] = mapped_column(init=True)
    stdout: Mapped[str] = mapped_column(init=True)
    stderr: Mapped[str] = mapped_column(init=True)
    cyst_run_id: Mapped[str] = mapped_column(init=True)
    status: Mapped[str] = mapped_column(init=True)

    run_id: Mapped[int] = mapped_column(ForeignKey("challenge_run_statistics.id"), init=True)
    run: Mapped[DBRun] = relationship(back_populates="episodes")


class RunStatus(Enum):
    """
    A status of a run or episode.

        :INIT: An initial status before the execution.
        :RUNNING: A run or episode is being executed.
        :FINISHED: A run or episode successfully finished.
        :ERROR: There was an error executing the run or episode.
    """
    INIT = auto()
    RUNNING = auto()
    FINISHED = auto()
    ERROR = auto()


@dataclass
class Episode:
    """
    A class tracking one episode of a run. You should never directly create this class instances.

    :param cyst_run_id: A CYST run ID for the episode.
    :param stdout: A copy of stdout contents.
    :param stderr: A copy of stderr contents.
    :param run: An ID of a run this episode belongs to.
    :param status: A status of the episode.
    """
    cyst_run_id: str
    stdout: str
    stderr: str
    run: int = -1
    status: RunStatus = RunStatus.RUNNING


@dataclass
class Run:
    """
    A class tracking a run. You should never directly create this class instances.

    :param specification: A run specification of the run.
    :param executor: A process pool that is used to execute the episodes.
    :param status: A status of the run.
    :param detail: Supplementary information for the run, usually error messages.
    :param running: A set of IDs of episodes that are currently running.
    :param successful: A set of IDs of episodes that successfully finished.
    :param error: A set of IDs of episodes that encountered an error.
    :param episodes: All episodes within the run.
    :param id: An ID of a run. Should be unique for the challenge instance.
    """
    specification: RunSpecification
    executor: Optional[ProcessPoolExecutor] = None
    status: RunStatus = RunStatus.INIT
    detail: str = ""
    running: Set[int] = field(default_factory=set)
    successful: Set[int] = field(default_factory=set)
    error: Set[int] = field(default_factory=set)
    episodes: Dict[int, Episode] = field(default_factory=dict)
    id: int = field(default=-1)



import inspect

from pathlib import Path
from typing import List, Dict

from aica_challenge_1.package_manager import PackageManager
from aica_challenge_1.scenario_manager import ScenarioManager
from aica_challenge_1.execution_manager import ExecutionManager, RunSpecification

from cyst.api.environment.environment import Environment

class Challenge:
    """
    This is the main class for programmatic manipulation with the AICA challenge. It provides access to challenge
    management functionality and enables launching of training runs.

    In most of the cases it is better to use the TUI.

    Example:

        from aica_challenge_1.challenge import Challenge

        # It is better to hide everything behind this flag if you are to ever execute
        # runs. They are separate processes and they could cause some strange errors.
        if __name__ == "__main__"
            challenge = Challenge()
    """
    def __init__(self, challenge_dir: str = ""):
        caller_frame = inspect.currentframe().f_back
        if "__file__" in caller_frame.f_locals:
            caller_dir = Path(caller_frame.f_locals["__file__"]).parent
        else:
            caller_dir = Path.cwd()

        if challenge_dir:
            cd = Path(challenge_dir)
            if not cd.is_absolute():
                cd = caller_dir / cd
        else:
            cd = caller_dir

        if not self.check_init_state(cd):
            raise ValueError("Challenge environment not properly initiated", cd)

        self._package_manager = PackageManager(cd / "agents")
        self._scenario_manager = ScenarioManager(cd / "scenarios")
        self._execution_manager = ExecutionManager(self._package_manager, self._scenario_manager)

    @property
    def packages(self) -> PackageManager:
        """
        Provides access to package and agent management utilities.

        :returns: package manager
        """
        return self._package_manager

    @property
    def scenarios(self) -> ScenarioManager:
        """
        Provides access to scenario management utilities.

        :returns: scenario manager
        """
        return self._scenario_manager

    @property
    def execution(self) -> ExecutionManager:
        """
        Provides access to run preparation and execution.

        :returns: execution manager
        """
        return self._execution_manager

    # Convenience function
    def execute(self, run_specification: RunSpecification | str, single_process=False) -> None:
        """
        Executes a run. This function is just a proxy for the same function in execution manager.

        :param run_specification: Either a RunSpecification object or a name of a run specification that is stored in
            the database.
        :param single_process: If set to True, it will execute only one run at a time (regardless of the run
            specification) and it will display the stdout and stderr. If set to False, it will execute each run in a
            new process (even if the run specification says no parallel runs) and stdout and stderr are hidden and
            stored in the database.
        """
        self._execution_manager.execute(run_specification, single_process)

    @staticmethod
    def init_environment():
        """
        Initializes the AICA challenge environment in the current working directory. The initialization currently
        entails only creation of `agents` and `scenarios` directories.
        """
        caller_frame = inspect.currentframe().f_back
        if "__file__" in caller_frame.f_locals:
            caller_dir = Path(caller_frame.f_locals["__file__"]).parent
        else:
            caller_dir = Path.cwd()

        agents_path = caller_dir / "agents"
        scenarios_path = caller_dir / "scenarios"

        if not agents_path.exists():
            print(f"Creating new agents directory: {str(agents_path)}")
            agents_path.mkdir()

        if not scenarios_path.exists():
            print(f"Creating new scenarios directory: {str(scenarios_path)}")
            scenarios_path.mkdir()

    @staticmethod
    def check_init_state(challenge_dir: Path) -> bool:
        """
        Checks, whether an AICA challenge is initialized at the specified directory. The function looks for a presence
        of `agents` and `scenarios` directories.

        :param challenge_dir: A directory where to check for the challenge environment.

        :returns: True if the challenge is initialized at the specified path.
        """
        if not challenge_dir.exists():
            return False

        agents_path = challenge_dir / "agents"
        scenarios_path = challenge_dir / "scenarios"

        if not (agents_path.exists() and scenarios_path.exists()):
            return False

        return True

    @staticmethod
    def list_actions() -> List[Dict[str, str]]:
        result = []

        env = Environment.create()
        actions = env.resources.action_store.get_prefixed("ac1")

        for action in actions:
            result.append({
                "id": action.id,
                "description": action.description,
                "parameters": [x.name for x in action.parameters.values()]
            })

        return result



import importlib.util
import json
import requests
import sys
import string
import secrets
import uuid

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

from cyst.api.configuration import ConfigItem


@dataclass
class ScenarioVariant:
    """
    Represents one variant of a particular scenario.

    :param id: An identifier of a this variant.
    :param path: A path to the variant's configuration file.
    :param config: A list of configuration items parsed from configuration file.
    """
    id: int
    path: str
    config: List[ConfigItem]


@dataclass
class Scenario:
    """
    Represents a scenario.

    :param name: A given name of the scenario.
    :param short_path: A name of the scenario directory inside the `scenarios` directory of the challenge.
    :param path: A full path to the scenario.
    :param description: A description of the scenario.
    :param goal: A goal of the scenario.
    :param parameters: A set of parameters that is passed to agents on top of the run parameters.
    :param variants: A list of variants for the scenario.
    """
    name: str
    short_path: str
    path: str
    description: str
    goal: str
    parameters: Dict[str, str]
    variants: Dict[int, ScenarioVariant]


class ScenarioManager:
    """
    Scenario manager takes care of tracking and downloading scenarios.
    """
    def __init__(self, scenarios_path: str | Path = ""):
        self._remote_scenarios: Dict[str, Scenario] = {}
        self._local_scenarios: Dict[str, Scenario] = {}

        self._repository_url = "https://gitlab.ics.muni.cz/aica/scenarios_1/-/raw/master/"

        if not scenarios_path:
            self._scenarios_path: Path = Path(__file__).parent / "scenarios"
        elif isinstance(scenarios_path, str):
            self._scenarios_path: Path = Path(scenarios_path)
        else:
            self._scenarios_path = scenarios_path

        if not self._scenarios_path.exists():
            raise ValueError("Path to scenarios does not exist.")

        self._scan_local_scenarios()

    @staticmethod
    def _get_configuration_objects(path: Path) -> List[ConfigItem]:
        module_name = "configuration_variant_" + str(uuid.uuid4())
        spec = importlib.util.spec_from_file_location(module_name, str(path))
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        configuration_objects =  module.configuration
        if type(configuration_objects) == List:
            return configuration_objects
        else:
            return [configuration_objects]

    def _scan_variant(self, id: str, variants_path: Path) -> Optional[ScenarioVariant]:
        path = variants_path / id

        configuration_path = path / "configuration.py"

        if not configuration_path.exists():
            print(f"Configuration files for the variant '{id}' not found. Variant ignored.")
            return None

        try:
            configuration_objects = self._get_configuration_objects(configuration_path)
        except:
            print(f"Failed to read the configuration file for the variant '{id}'. Variant ignored.")
            return None

        return ScenarioVariant(int(id), str(configuration_path), configuration_objects)

    def _scan_scenario(self, name: str, scenarios_path: Path) -> Optional[Scenario]:
        path = scenarios_path / name

        name_path = path / "name.md"
        if name_path.exists():
            name_s = name_path.read_text()
        else:
            name_s = name

        description_path = path / "description.md"
        if description_path.exists():
            description = description_path.read_text()
        else:
            description = "No description provided"

        goal_path = path / "goal.md"
        if goal_path.exists():
            goal = goal_path.read_text()
        else:
            goal = "No goal provided"

        parameters_path = path / "parameters.json"
        if parameters_path.exists():
            parameters = json.loads(parameters_path.read_text())
        else:
            parameters = {}

        result = Scenario(name_s, name, str(path), description, goal, parameters, {})

        variants_path = path / "variants"
        variants_on_path = set([x.parts[-1] for x in variants_path.iterdir() if x.is_dir()])

        for variants_id in variants_on_path:
            variant = self._scan_variant(variants_id, variants_path)
            if variant:
                result.variants[int(variants_id)] = variant

        if not result.variants:
            print(f"Could not find any valid scenario variants. Scenario '{name}' not included.")
            return None
        else:
            return result

    def _scan_local_scenarios(self) -> None:
        scenarios_on_path = set([x.parts[-1] for x in self._scenarios_path.iterdir() if x.is_dir()])

        for scenario_name in scenarios_on_path:
            scenario = self._scan_scenario(scenario_name, self._scenarios_path)
            if scenario:
                self._local_scenarios[scenario_name] = scenario

    def _scan_remote_scenarios(self) -> None:
        r = requests.get(self._repository_url + "scenarios.json")
        scenarios_info = r.json()

        for scenario in scenarios_info:
            if "parameters" in scenario:
                parameters = scenario["parameters"]
            else:
                parameters = {}

            s = Scenario(scenario["name"], scenario["path"], scenario["path"], scenario["description"], scenario["goal"], parameters, {})
            for variant in scenario["variants"]:
                v = ScenarioVariant(variant, str(variant), [])
                s.variants[variant] = v
            self._remote_scenarios[s.name] = s

    def download_remote_scenarios(self, overwrite: bool = False) -> None:
        """
        Attempts to download all scenarios that are currently published.

        :param overwrite: If False, it will not attempt to overwrite local scenarios. This is mostly useful in
            situations when you made local modifications that you want to preserve.
        """
        self._scan_remote_scenarios()

        for scenario in self._remote_scenarios.values():
            scenario_path = self._scenarios_path / scenario.path
            if not scenario_path.exists():
                scenario_path.mkdir()

            (scenario_path / "name.md").write_text(scenario.name)
            (scenario_path / "description.md").write_text(scenario.description)
            (scenario_path / "goal.md").write_text(scenario.goal)
            (scenario_path / "parameters.json").write_text(json.dumps(scenario.parameters))

            for variant in scenario.variants.values():
                variant_path_remote = Path(scenario.path) / variant.path / "configuration.py"
                variant_path_local = self._scenarios_path / scenario.path / "variants" / variant.path / "configuration.py"
                variant_config = requests.get(self._repository_url + variant_path_remote.as_posix()).text
                if not variant_path_local.parent.exists():
                    variant_path_local.parent.mkdir(parents=True)
                    variant_path_local.write_text(variant_config)
                elif overwrite:
                    variant_path_local.write_text(variant_config)

    def get_scenario(self, name: str) -> Optional[Scenario]:
        """
        Retrieve a local scenario by name.

        :param name: A short name (i.e., the directory in `scenarios` directory) of the scenario.

        :return: A scenario description if there is a scenario of the given name, None otherwise.
        """
        if name not in self._local_scenarios:
            return None
        else:
            return self._local_scenarios[name]

    def get_scenarios(self, remote: bool = False) -> List[Scenario]:
        """
        Retrieves descriptions of either local or remote scenarios.

        :param remote: If set to True, checks for all available scenarios online. Otherwise look only inside the
            `scenarios` directory.

        :return: A list of scenario descriptions.
        """
        if remote:
            self._scan_remote_scenarios()
            return list(sorted(self._remote_scenarios.values(), key=lambda v: v.name))
        else:
            self._scan_local_scenarios()
            return list(sorted(self._local_scenarios.values(), key=lambda v: v.name))



import asyncio

from cyst.api.logic.access import AccessLevel
from cyst.api.logic.exploit import ExploitCategory, ExploitLocality
from netaddr import IPNetwork
from sqlalchemy import create_engine, select
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, MappedAsDataclass, Session, relationship
from typing import Tuple, Callable, Union, List, Coroutine, Optional

from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.message import Request, Response, Status, StatusOrigin, StatusValue, StatusDetail
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.platform_specification import PlatformSpecification, PlatformType
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.logic.action import ActionDescription, ActionParameterType, ActionParameter, Action, ActionType
from cyst.api.logic.behavioral_model import BehavioralModel, BehavioralModelDescription
from cyst.api.logic.composite_action import CompositeActionManager
from cyst.api.network.node import Node
from cyst.api.utils.duration import Duration, msecs


class AC1Model(BehavioralModel):
    def __init__(self, configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                 policy: EnvironmentPolicy, messaging: EnvironmentMessaging, infrastructure: EnvironmentInfrastructure,
                 composite_action_manager: CompositeActionManager) -> None:

        self._configuration = configuration
        self._action_store = resources.action_store
        self._exploit_store = resources.exploit_store
        self._policy = policy
        self._messaging = messaging
        self._infrastructure = infrastructure
        self._cam = composite_action_manager

        self._action_store.add(ActionDescription(
            id="ac1:inspect",
            type=ActionType.DIRECT,
            platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"),
                      PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
            description="Inspecting a machine, where you have an access to.",
            parameters=[])
        )

        self._action_store.add(ActionDescription(
            id="ac1:scan_host",
            type=ActionType.DIRECT,
            platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"),
                      PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
            description="Scanning of one specific host",
            parameters=[])
        )

        self._action_store.add(ActionDescription(
            id="ac1:scan_network",
            type=ActionType.COMPOSITE,
            platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"),
                      PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
            description="Scanning of a subnet",
            parameters=[ActionParameter(type=ActionParameterType.NONE, name="net",
                                        domain=configuration.action.create_action_parameter_domain_any())])
        )

        self._action_store.add(ActionDescription(
            id="ac1:access_target",
            type=ActionType.DIRECT,
            platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"),
                      PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
            description="Use either exploit or a valid authentication/authorization to open a session to a target service",
            parameters=[])
        )

        self._action_store.add(ActionDescription(
            id="ac1:exfiltrate_data",
            type=ActionType.DIRECT,
            platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"),
                      PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
            description="Exfiltrate an interesting data from a target.",
            parameters=[ActionParameter(type=ActionParameterType.NONE, name="path",
                                        domain=configuration.action.create_action_parameter_domain_any())]
        ))

    async def action_flow(self, message: Request) -> Tuple[Duration, Response]:
        if not message.action:
            raise ValueError("Action not provided")

        action_name = "_".join(message.action.fragments)
        fn: Callable[[Request], Coroutine[None, None, Tuple[Duration, Response]]] = getattr(self, "process_" + action_name, self.process_default_flow)
        duration, response = await fn(message)

        return duration, response

    async def action_effect(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        if not message.action:
            raise ValueError("Action not provided")

        action_name = "_".join(message.action.fragments)
        fn: Callable[[Request, Node], Coroutine[None, None, Tuple[Duration, Response]]] = getattr(self, "process_" + action_name, self.process_default_effect)
        duration, response = await fn(message, node)

        return duration, response

    def action_components(self, message: Union[Request, Response]) -> List[Action]:
        return []

    async def process_default_effect(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        error = f"A direct action with unknown semantics specified: {message.action.id}."
        return msecs(0), self._messaging.create_response(message, status=Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                         content=error, session=message.session)

    async def process_default_flow(self, message: Request) -> Tuple[Duration, Response]:
        error = f"A composite action with unknown semantics specified: {message.action.id}."
        return msecs(0), self._messaging.create_response(message, status=Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                         content=error, session=message.session)

    async def process_inspect(self, message: Request, node: Node) -> Tuple[Duration, Response]:

        if message.dst_ip != "127.0.0.1" and message.src_ip != message.dst_ip and \
                (message.session and message.session.end[0] != message.dst_ip):

            error = f"The agent does not have an access to the following IP address {message.dst_ip}."
            return msecs(20), self._messaging.create_response(message,
                                                              status=Status(StatusOrigin.SERVICE, StatusValue.FAILURE),
                                                              content=error, session=message.session)

        result = None
        status = None

        if not message.dst_service:
            services = []
            for service in node.services.values():
                if service.passive_service:
                    services.append((service.name, str(service.passive_service.version)))

            result = {
                "ips": [str(x) for x in node.ips],
                "services": services
            }
            status = Status(StatusOrigin.SERVICE, StatusValue.SUCCESS)
        else:
            if not message.dst_service in node.services:
                result = f"The service '{message.dst_service}' does not exist on the target machine."
                status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.SERVICE_NOT_EXISTING)
            else:
                service = node.services[message.dst_service].passive_service
                result = {
                    "name": message.dst_service,
                    "version": service.version,
                    "data": [x.path for x in self._configuration.service.private_data(service)]
                }
                status = Status(StatusOrigin.SERVICE, StatusValue.SUCCESS)

        return msecs(20), self._messaging.create_response(message, status=status, content=result,
                                                          session=message.session, auth=message.auth)

    async def process_scan_host(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        host_ip = message.dst_ip
        services = []
        for service in node.services.values():
            if service.passive_service and not service.passive_service.local:
                services.append((service.name, str(service.passive_service.version)))

        result = {
            "ip": host_ip,
            "services": services
        }

        return msecs(40), self._messaging.create_response(message,
                                                          status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                          session=message.session, auth=message.auth, content=result)

    async def process_scan_network(self, message: Request) -> Tuple[Duration, Response]:
        if "net" not in message.action.parameters or not message.action.parameters["net"].value:
            return msecs(0), self._messaging.create_response(message,
                                                             status=Status(StatusOrigin.SERVICE, StatusValue.FAILURE,
                                                                           StatusDetail.ACTION_PARAMETER_NOT_PROVIDED),
                                                             session=message.session, auth=message.auth)

        net = IPNetwork(message.action.parameters["net"].value)
        if not net:
            return msecs(0), self._messaging.create_response(message,
                                                             status=Status(StatusOrigin.SERVICE, StatusValue.FAILURE,
                                                                           StatusDetail.ACTION_PARAMETER_NOT_APPLICABLE),
                                                             session=message.session, auth=message.auth)

        tasks = []

        if net.prefixlen < 31:
            ips = [net.cidr.ip]
            ips.extend(list(net.iter_hosts()))
            ips.append(net.broadcast)
        else:
            ips = [net.iter_hosts()]

        for ip in ips:
            action = self._action_store.get("ac1:scan_host")
            request = self._messaging.create_request(ip, "", action, original_request=message)
            tasks.append(self._cam.call_action(request, 0))

        results: List[Response] = await asyncio.gather(*tasks)
        successes = []
        failures = []
        errors = []
        for r in results:
            if r.status.value == StatusValue.SUCCESS:
                successes.append(r.content)
            elif r.status.value == StatusValue.FAILURE:
                failures.append(r.src_ip)
            elif r.status.value == StatusValue.ERROR:
                errors.append(r.src_ip)

        content = {
            "success": successes,
            "failure": failures,
            "error": errors
        }

        response = self._messaging.create_response(message,
                                                   status=Status(StatusOrigin.NETWORK, StatusValue.SUCCESS),
                                                   content=content, session=message.session, auth=message.auth)

        return msecs(0), response

    async def process_access_target(self, message: Request, node: Node) -> Tuple[Duration, Response]:

        content = ""
        status = None
        session = None

        # A target service is provided
        if not message.dst_service:
            content = "The action requires a specific service as a target."
            status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.SERVICE_NOT_PROVIDED)

        # The service exists at the target
        if not content and message.dst_service not in node.services:
            content = f"Service '{message.dst_service}' is not running at the target."
            status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.SERVICE_NOT_EXISTING)

        # The service can be accessed
        if not content:
            target_service = node.services[message.dst_service].passive_service
            if target_service.local and message.src_ip not in node.ips:
                if not message.session or message.session.end[0] not in node.ips:
                    content = f"Service '{message.dst_service}' can only be accessed locally."
                    status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.SESSION_NOT_PROVIDED)

        # Either exploit or auth was provided
        if not content:
            if not message.action.exploit and not message.auth:
                content = f"The action requires either an exploit or a correct authorization"
                status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.EXPLOIT_NOT_PROVIDED | StatusDetail.AUTHORIZATION_NOT_PROVIDED | StatusDetail.AUTHENTICATION_NOT_PROVIDED)

        # Exploit is being used
        if not content and message.action.exploit:
            # It is the correct category
            if message.action.exploit.category != ExploitCategory.CODE_EXECUTION:
                content = "Provided exploit cannot be used to open a session."
                status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.EXPLOIT_CATEGORY_NOT_APPLICABLE)
            # It is the correct locality
            elif message.action.exploit.locality == ExploitLocality.LOCAL:
                if message.src_ip not in node.ips or (message.session and message.session.end[0] not in node.ips):
                    content = "Attempting to use a local exploit remotely."
                    status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.EXPLOIT_LOCALITY_NOT_APPLICABLE)
            # It can be used on the target service
            elif not self._exploit_store.evaluate_exploit(message.action.exploit, message, node):
                content = "The provided exploit cannot be used"
                status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.EXPLOIT_NOT_APPLICABLE)

            content = "Session successfully opened"
            status = Status(StatusOrigin.SERVICE, StatusValue.SUCCESS)
            session = self._messaging.open_session(message)

        # Auth is being used (Authentication shoudl be handled before entering the behavioral model)
        if not content and message.auth:
            result, reason = self._policy.decide(node, message.dst_service, AccessLevel.LIMITED, message.auth)
            if not result:
                content = f"Could not use the authorization to open a session. Reason: {reason}"
                status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.AUTHORIZATION_NOT_APPLICABLE)
            else:
                content = "Session successfully opened"
                status = Status(StatusOrigin.SERVICE, StatusValue.SUCCESS)
                session = self._messaging.open_session(message)

        response = self._messaging.create_response(message, status, content, session if session else message.session, message.auth)
        return msecs(20), response

    async def process_exfiltrate_data(self, message: Request, node: Node) -> Tuple[Duration, Response]:

        content = ""
        status = None

        # A target service is provided
        if not message.dst_service:
            content = "The action requires a specific service as a target."
            status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.SERVICE_NOT_PROVIDED)

        # The service exists at the target
        if not content and message.dst_service not in node.services:
            content = f"Service '{message.dst_service}' is not running at the target."
            status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.SERVICE_NOT_EXISTING)

        # The path is provided
        if not content and "path" not in message.action.parameters:
            content = "Required parameter 'path' not provided."
            status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.ACTION_PARAMETER_NOT_PROVIDED)

        if not content:
            service = node.services[message.dst_service].passive_service
            service_data = {x.path: x for x in self._configuration.service.private_data(service)}
            data_path = message.action.parameters["path"].value

            # local extraction
            if message.src_ip in node.ips or (message.session and message.session.end[0] in node.ips):
                if data_path not in service_data:
                    content = f"Data with required path '{data_path}' not within the service."
                    status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.ACTION_PARAMETER_NOT_APPLICABLE)
                else:
                    content = {"path": data_path, "content": service_data[data_path].description}
                    status = Status(StatusOrigin.SERVICE, StatusValue.SUCCESS)
            # remote extraction
            else:
                if not message.action.exploit:
                    content = "An exploit is required for remote data exfiltration."
                    status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.EXPLOIT_NOT_PROVIDED)

                if not content and message.action.exploit.category != ExploitCategory.DATA_MANIPULATION:
                    content = "The exploit does not have a suitable DATA_MANIPULATION category."
                    status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.EXPLOIT_CATEGORY_NOT_APPLICABLE)

                if not content and message.action.exploit.locality != ExploitLocality.REMOTE:
                    content = "Attempting to use a local exploit remotely."
                    status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.EXPLOIT_LOCALITY_NOT_APPLICABLE)

                if not content and not self._exploit_store.evaluate_exploit(message.action.exploit, message, node):
                    content = "The provided exploit cannot be used"
                    status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.EXPLOIT_NOT_APPLICABLE)

                if data_path not in service_data:
                    content = f"Data with required path '{data_path}' not within the service."
                    status = Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.ACTION_PARAMETER_NOT_APPLICABLE)
                else:
                    content = {"path": data_path, "content": service_data[data_path].description}
                    status = Status(StatusOrigin.SERVICE, StatusValue.SUCCESS)

        response = self._messaging.create_response(message, status, content, message.session, message.auth)
        return msecs(20), response


def create_ac1_model(configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                      policy: EnvironmentPolicy, messaging: EnvironmentMessaging,
                      infrastructure: EnvironmentInfrastructure,
                      composite_action_manager: CompositeActionManager) -> BehavioralModel:
    model = AC1Model(configuration, resources, policy, messaging, infrastructure, composite_action_manager)
    return model


behavioral_model_description = BehavioralModelDescription(
    namespace="ac1",
    description="Behavioral model for the first AICA challenge.",
    creation_fn=create_ac1_model,
    platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")]
)



import functools
import io
import os

from concurrent.futures import ProcessPoolExecutor, Future
from contextlib import redirect_stderr, redirect_stdout, nullcontext
from dataclasses import dataclass, field
from enum import Enum, auto
from random import choice
from sqlalchemy import create_engine, select, ForeignKey, delete
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, MappedAsDataclass, Session, relationship
from typing import Dict, Any, Optional, List, Self, Set, Tuple

from cyst.api.environment.environment import Environment
from cyst.api.utils.counter import Counter

from aica_challenge_1.package_manager import PackageManager
from aica_challenge_1.scenario_manager import ScenarioManager, ScenarioVariant
from aica_challenge_1.launcher import launch_simulation
from aica_challenge_1.execution_structures import Episode, DBEpisode, DBRun, RunStatus, RunSpecification, Run, Base, \
    DBRunParameter, DBRunSpecification


class ExecutionManager:
    """
    The execution manager retrieves run specifications and takes care of run executions.
    """
    def __init__(self, package_manager: PackageManager, scenario_manager: ScenarioManager):
        self._package_manager = package_manager
        self._scenario_manager = scenario_manager
        self._run_specifications: Dict[str, RunSpecification] = {}
        self._run_specifications_by_id: Dict[id, RunSpecification] = {}

        self._db = create_engine("sqlite+pysqlite:///aica_challenge.db")
        Base.metadata.create_all(self._db)

        with Session(self._db) as session:
            for obj in session.execute(select(DBRunSpecification)).scalars():
                spec = RunSpecification.from_db_spec(obj)
                self._run_specifications[obj.name] = spec
                self._run_specifications_by_id[obj.id] = spec

            parameters: Dict[int, List[DBRunParameter]] = {}
            for obj in session.execute(select(DBRunParameter)).scalars():
                if not obj.run_id in parameters:
                    parameters[obj.run_id] = []
                parameters[obj.run_id].append(obj)

            for spec in self._run_specifications.values():
                if spec.db_id in parameters:
                    for param in parameters[spec.db_id]:
                        spec.parameters[param.key] = param.value

        self._runs: Dict[int, Run] = {}

    def list_run_specifications(self) -> List[str]:
        """
        Provides a list of run names that are available for execution.

        :return: A list of run names.
        """
        return sorted(self._run_specifications.keys())

    def get_run_specification(self, name: str) -> Optional[RunSpecification]:
        """
        Attempts to retrieve a run specification by name.

        :param name: A name of the specification.

        :return: A run specification if it exists for a given name, or None otherwise.
        """
        return self._run_specifications.get(name, None)

    def set_run_specification(self, specification: RunSpecification, old_specification: Optional[RunSpecification] = None) -> None:
        """
        Saves a run specification to the database.

        :param specification: A specification that should be saved.
        :param old_specification: If provided, this specification will be overwritten by the other specification
        """
        if old_specification and old_specification.name and old_specification.name != specification.name:
            del self._run_specifications[str(old_specification.name)]

        self._run_specifications[specification.name] = specification

        with Session(self._db) as session:
            db_obj = None
            if specification.db_id != -1:
                db_obj = session.execute(select(DBRunSpecification).where(DBRunSpecification.id==specification.db_id)).scalar_one()
                db_obj.replace(specification)
            else:
                db_obj = DBRunSpecification.copy(specification)

            session.add(db_obj)
            session.flush()

            specification_keys = set(specification.parameters.keys())
            db_keys = set()

            for obj in session.execute(select(DBRunParameter).where(DBRunParameter.run_id==db_obj.id)).scalars():
                # Key was deleted
                if obj.key not in specification.parameters:
                    session.execute(delete(obj))
                else:
                    db_keys.add(obj.key)
                    obj.value = specification.parameters[obj.key]
                    session.add(obj)

            new_keys = specification_keys - db_keys
            for key in new_keys:
                session.add(DBRunParameter(run_id=db_obj.id, key=key, value=specification.parameters[key]))

            session.commit()

    def save_run_information(self, run: Run) -> None:
        with Session(self._db, expire_on_commit=False) as session:
            stmt = select(DBRun).where(DBRun.id == run.id)
            run_db: DBRun = session.scalars(stmt).one()

            details = f"Successfull episodes: {sorted(list(run.successful))}, failed episodes: {sorted(list(run.error))}"

            run.detail = details
            run_db.details = details

            # We set status to finished whenever there was at elast one successful episode
            status = RunStatus.FINISHED if run.successful else RunStatus.ERROR
            run.status = status
            run_db.status = str(status)

            session.commit()

    def get_run(self, run_id: int) -> Run:
        """
        Gets the information about a specific run.

        :param run_id: An ID of the run to get.

        :return: A run information.
        """
        return self._runs[run_id]

    def get_runs(self) -> List[Tuple[int, RunStatus]]:
        """
        Gets the IDs and statuses of all runs executed in the challenge instance.

        :return: Tuple containing the ID [0] and status [1] for each run.
        """
        result = []
        for k in sorted(self._runs.keys()):
            r = self._runs[k]
            result.append((k, self._runs[k].status))
        return result

    def execution_callback(self, episode_number: int, future: Future):
        if future.exception():
            print(f"The episode {episode_number} has been terminated.")
            return

        e: Episode = future.result()
        run = self._runs[e.run]
        run.episodes[episode_number] = e
        run.running.remove(episode_number)
        if e.status == RunStatus.FINISHED:
            run.successful.add(episode_number)
        else:
            run.error.add(episode_number)

        if not run.running:
            self.save_run_information(run)

    def execute(self, specification: RunSpecification | str, single_process=False) -> None:
        """
        Executes a run.

        :param specification: Either a RunSpecification object or a name of a run specification that is stored in
            the database.
        :param single_process: If set to True, it will execute only one run at a time (regardless of the run
            specification) and it will display the stdout and stderr. If set to False, it will execute each run in a
            new process (even if the run specification says no parallel runs) and stdout and stderr are hidden and
            stored in the database.
        """
        if isinstance(specification, str):
            spec_name = specification
            specification = self._run_specifications.get(specification, None)
            if not specification:
                raise ValueError(f"Run with the name '{spec_name}' not available in the system.")

        with Session(self._db) as session:
            """
            # Refresh run specification
            specification = session.get(DBRunSpecification, specification.db_id)

            if not specification:
               raise ValueError(f"There was an error extracting run specification from the database")

            parameters = []
            for obj in session.execute(select(DBRunParameter).where(DBRunParameter.run_id==specification.id)).scalars():
                parameters.append(obj)
            """

            run = Run(specification)

            db_specification = session.get(DBRunSpecification, specification.db_id)
            db_run = DBRun(
                status=str(RunStatus.INIT),
                details="",
                episodes=[],
                specification_id=specification.db_id,
                specification=db_specification
            )

            session.add(db_run)
            session.flush()

            if not specification.name:
                run.status = RunStatus.ERROR
                run.detail = "Run specification must have a name."
                db_run.status = str(RunStatus.ERROR)
                db_run.details = run.detail
                session.commit()
                return

            if specification.agent_name not in self._package_manager.list_installed_agents():
                run.status = RunStatus.ERROR
                run.detail = f"Chosen agent '{specification.agent_name}' not installed in the system."
                db_run.status = str(RunStatus.ERROR)
                db_run.details = run.detail
                session.commit()
                return

            scenario_name = specification.scenario
            scenario = self._scenario_manager.get_scenario(scenario_name)

            if scenario_name == "Random":
                scenarios = self._scenario_manager.get_scenarios()
                if not scenarios:
                    run.status = RunStatus.ERROR
                    run.detail = "No scenarios installed in the system. Cannot choose a random one."
                    db_run.status = str(RunStatus.ERROR)
                    db_run.details = run.detail
                    session.commit()
                    return
                s = choice(scenarios)
                scenario_name = s.short_path
                scenario = s

            if not scenario:
                run.status = RunStatus.ERROR
                run.detail = f"Chosen scenario '{specification.scenario}' not available."
                db_run.status = str(RunStatus.ERROR)
                db_run.details = run.detail
                session.commit()
                return

            variant_id = specification.variant
            variants = scenario.variants

            if variant_id == -1:
                if not variants:
                    run.status = RunStatus.ERROR
                    run.detail = f"No variants of the scenario '{scenario_name}' exist. Cannot choose a random one."
                    db_run.status = str(RunStatus.ERROR)
                    db_run.details = run.detail
                    session.commit()
                    return
                variant_id = choice(list(variants.keys()))

            elif variant_id not in variants:
                run.status = RunStatus.ERROR
                run.detail = f"Variant '{variant_id}' of the scenario '{scenario_name}' is not available in the system."
                db_run.status = str(RunStatus.ERROR)
                db_run.details = run.detail
                session.commit()
                return

            scenario_variant = self._scenario_manager.get_scenario(scenario_name).variants[variant_id]

            run.id = db_run.id
            self._runs[run.id] = run

            run.status = RunStatus.RUNNING
            db_run.status = str(RunStatus.RUNNING)

            session.commit()

        # Add scenario-level parameters unless overwritten by the user
        for key, value in scenario.parameters.items():
            if key not in specification.parameters:
                specification.parameters[key] = value

        if not single_process:
            run.executor = ProcessPoolExecutor(max_workers=specification.max_parallel)

            for e in range(specification.max_episodes):
                future: Future = run.executor.submit(launch_simulation, run.id, e, scenario_variant,
                                                     specification.agent_name, specification.max_time,
                                                     specification.max_actions, specification.parameters)
                future.add_done_callback(functools.partial(self.execution_callback, e))
                run.running.add(e)
        else:
            for e in range(specification.max_episodes):
                ep = launch_simulation(run.id, e, scenario_variant, specification.agent_name, specification.max_time,
                                       specification.max_actions, specification.parameters, supress_output=False)

                if ep.status == RunStatus.FINISHED:
                    run.successful.add(e)
                else:
                    run.error.add(e)

            self.save_run_information(run)
